local B = CBW_Battle
local CV = B.Console
local CP = B.ControlPoint

B.GetInteractionType = function(smo,tmo)
	//0 = No interaction. 1 = Bump. 2 = (unfinished) 3 = Full damage code
	if not(smo and smo.valid and smo.health and smo.player and not smo.player.spectator)
		or not(tmo and tmo.valid and tmo.health and tmo.player and not tmo.player.spectator)
		then return 0
	end
	//Collision FriendlyFire
	if CV.Collision.value and CV_FindVar("friendlyfire").value return 3 end
	//Egg Robo Tag
	if gametype == GT_EGGROBOTAG return 1 end
	//Battle gametype
	if B.BattleGametype() and not(B.MyTeam(smo.player,tmo.player)) then return 3 end
	//Tag with collision
	if G_TagGametype() and CV.Collision.value return 1 end
	//Ringslinger with collision
	if G_RingSlingerGametype() and CV.Collision.value and not(B.MyTeam(smo.player,tmo.player)) return 3 end
	//Platforming with collision
	if CV.Collision.value return 1 end
	//Default
	return 0
end

B.GetCollideRelativeAngle = function(mo,collide,drawangle)
	if not (mo and mo.valid and mo.health and mo.player and collide and collide.valid and collide.health)
		then return nil
	end
	local collideangle
	if mo.player and drawangle then
		collideangle = mo.player.drawangle - R_PointToAngle2(mo.x-mo.momx,mo.y-mo.momy,collide.x-collide.momx,collide.y-collide.momy)
	else
		collideangle = mo.angle - R_PointToAngle2(mo.x-mo.momx,mo.y-mo.momy,collide.x-collide.momx,collide.y-collide.momy)
	end
	if collideangle >= 180*ANG1 then
		collideangle = $- 360*ANG1
	elseif collideangle <= -180*ANG1 then
		collideangle = $+360*ANG1
	end
	return collideangle
end

B.GetZCollideAngle = function(mo,collide)
	if not(mo and mo.valid and mo.health and collide and collide.valid and collide.health)	
		then 
		print("\x82 Error:\x80 An object in GetZCollideAngle() does not exist!")
		return 0
	end
	local x = mo.x - mo.momx
	local y = mo.y - mo.momy
	local z = mo.z+mo.height/2 - mo.momz
	local cx = collide.x - collide.momx
	local cy = collide.y - collide.momy
	local cz = collide.z+collide.height/2 - collide.momz
	local zdist = z-cz
	local xydist = FixedHypot(x-cx,y-cy)
	local collideangle = R_PointToAngle2(0,0,xydist,zdist)
	return collideangle
end

B.DoPlayerCollisionDamage = function(smo,tmo)
	local sbias = 0 
	local tbias = 0
	local satk = smo.player.battle_atk
	local tatk = tmo.player.battle_atk
	local sdef = smo.player.battle_def
	local tdef = tmo.player.battle_def
	local sbias = sdef-tatk // 0 == uncurl. -1 == take damage
	local tbias = tdef-satk
	
	local spower = (smo.player.powers[pw_super] or smo.player.powers[pw_invulnerability])
	local tpower = (tmo.player.powers[pw_super] or tmo.player.powers[pw_invulnerability])
	local stagit = smo.player.pflags&PF_TAGIT
	local ttagit = tmo.player.pflags&PF_TAGIT
	local sinvuln = smo.player.powers[pw_flashing]
	local tinvuln = tmo.player.powers[pw_flashing]
	if spower and not(tpower or tinvuln) then P_DamageMobj(tmo,smo,smo,0) return 3
	elseif tpower and not(spower or sinvuln) then P_DamageMobj(smo,tmo,tmo,0) return -3
	else
		if sbias < 0 and not(sinvuln or spower) then P_DamageMobj(smo,tmo,tmo,0) end
		if tbias < 0 and not(tinvuln or tinvuln) then P_DamageMobj(tmo,smo,smo,0) end
	end
	return sbias-tbias //Used for recoil and knockback distribution
end

B.UpdateRecoilState = function(mo)
	if not(mo and mo.valid and mo.health and mo.player) then return nil //Object invalid
	elseif mo.player.powers[pw_nocontrol] > 0 and mo.recoilangle != nil and mo.recoilthrust != nil then
		if P_PlayerInPain(mo.player) then
			mo.recoilangle = nil
			mo.recoilthrust = nil
			return false
		end
		if P_IsObjectOnGround(mo) then //Grounded
			if not(mo.player.charability == CA_GLIDEANDCLIMB) then
				mo.state = S_PLAY_SKID
				mo.player.panim = 0
				mo.frame = 0
			else
				mo.state = S_PLAY_GLIDE_LANDING
				mo.player.powers[pw_nocontrol] = $-1
			end
			if mo.player.skidtime == 0 and mo.recoilthrust > mo.scale*5 then
				S_StartSound(mo, sfx_skid)
			end
			mo.player.powers[pw_nocontrol] = min($,TICRATE)
			mo.player.skidtime = mo.player.powers[pw_nocontrol]
			//Apply recoil thrust
			P_InstaThrust(mo,mo.recoilangle,mo.recoilthrust)
			//Add friction
			mo.recoilthrust = ($+FixedMul($,FixedMul(mo.friction,mo.movefactor)))/2
		else //Aerial
			//Keep our variables up with current trajectory
			mo.recoilthrust = FixedHypot(mo.momx,mo.momy)
			mo.recoilangle = R_PointToAngle2(0,0,mo.momx,mo.momy)
			mo.player.panim = PA_FALL
			mo.state = S_PLAY_FALL
		end
		mo.player.drawangle = mo.recoilangle+ANGLE_180
		return true
	else //Not in disadvantage state, reset variables
		mo.recoilangle = nil
		mo.recoilthrust = nil
		return false
	end
end

B.DoPlayerInteract = function(smo,tmo)
	if not(smo and tmo and smo.valid and tmo.valid and smo.health and tmo.health)
		or (smo.pushed == nil or tmo.pushed == nil)
		or ((smo.pushed_last == smo.pushed) or (tmo.pushed_last == tmo.pushed))
		then return false
	end
	local splr = smo.player
	local tplr = tmo.player
	splr.pushed_creditplr = tplr
	splr.pushed_credittime = TICRATE*3
	tplr.pushed_creditplr = splr
	tplr.pushed_credittime = TICRATE*3
	//Get current attack and defensive priorities
	B.DoPriority(splr)
	B.DoSPriority(splr,tmo)
	local s_atk = splr.battle_atk
	local s_def = splr.battle_def
	B.DoPriority(tplr)
	B.DoSPriority(tplr,smo)
	local t_atk = tplr.battle_atk
	local t_def = tplr.battle_def
	
	local s_recoilfactor = FRACUNIT*(t_def-s_def+4)/8
	local t_recoilfactor = FRACUNIT*(s_def-t_def+4)/8
	
	local collisiontype = B.GetInteractionType(smo,tmo)
	
	//No more homing
-- 	splr.homing = 0
-- 	tplr.homing = 0
	
	//Get thrust
	local sspd = FixedHypot(smo.momx,smo.momy)
	local tspd = FixedHypot(tmo.momx,tmo.momy)
	local szspd = abs(smo.momz)
	local tzspd = abs(tmo.momz)
	local sangle = R_PointToAngle2(tmo.x-tmo.momx,tmo.y-tmo.momy,smo.x-smo.momx,smo.y-smo.momy)
	local tangle = R_PointToAngle2(smo.x-smo.momx,smo.y-smo.momy,tmo.x-tmo.momx,tmo.y-tmo.momy)
	local soundthrust = min(tmo.scale,smo.scale)
	local sthrust = FixedMul(tspd*2/3,tmo.scale-smo.scale/2)+smo.scale
	local sthrust2 = FixedMul(sspd+sspd/3,tmo.scale-smo.scale/2)+smo.scale
	local tthrust = FixedMul(sspd*2/3,smo.scale-tmo.scale/2)+tmo.scale
	local tthrust2 = FixedMul(tspd+tspd/3,smo.scale-tmo.scale/2)+tmo.scale
	local smomz = smo.momz
	local tmomz = tmo.momz

	
	//Get collision angle
	local tcollideangle = B.GetCollideRelativeAngle(tmo,smo)
	local scollideangle = B.GetCollideRelativeAngle(smo,tmo)
	
	local tzcollideangle = B.GetZCollideAngle(tmo,smo)
	local szcollideangle = B.GetZCollideAngle(smo,tmo)
	local tpriority = 0
	local spriority = 0
	local tprioritytype = 0
	local sprioritytype = 0
	
	local bias = 0
	//Collision info
	if collisiontype <= 1 and G_TagGametype() and not(leveltime < 30*TICRATE) then //Battle Tag
		if tplr.pflags&PF_TAGIT and not(splr.pflags&PF_TAGIT) and B.PlayerCanBeDamaged(splr) then
			splr.pflags = $|PF_TAGIT
			P_DoPlayerPain(smo.player, tmo, tmo)
			P_AddPlayerScore(tmo.player,100)
			if multiplayer and not(splitscreen) then print(B.CustomHurtMessage(smo.player,tmo,"tagging hand")) end
		elseif splr.pflags&PF_TAGIT and not(tplr.pflags&PF_TAGIT) and B.PlayerCanBeDamaged(tplr) then
			tplr.pflags = $|PF_TAGIT
			P_DoPlayerPain(tmo.player, smo, smo)
			P_AddPlayerScore(smo.player,100)
			if multiplayer and not(splitscreen) then print(B.CustomHurtMessage(tmo.player,smo,"tagging hand")) end
		end
	elseif collisiontype > 1 then //Standard Battle rules
		//Get collision prioritytype
-- 		tprioritytype = B.GetAttackPriorityType(tmo)
-- 		sprioritytype = B.GetAttackPriorityType(smo)
		
		//Update attack priority
-- 		tpriority = B.GetAttackPriority(tprioritytype,tcollideangle,tzcollideangle)
-- 		spriority = B.GetAttackPriority(sprioritytype,scollideangle,szcollideangle)
		
		//Do collision damage
		bias = B.DoPlayerCollisionDamage(smo,tmo)
	end
	
	local spain = (splr.powers[pw_nocontrol] > 0 or splr.panim == PA_PAIN or smo.health < 1)
	local tpain = (tplr.powers[pw_nocontrol] > 0 or tplr.panim == PA_PAIN or tmo.health < 1)
	
	//Apply Thrust
	if smo.health then
		P_Thrust(smo,sangle,max(tmo.scale,sthrust))
		if not spain then P_Thrust(smo,tangle+ANGLE_180,max(tmo.scale,sthrust2)) end
	end
	if tmo.health then
		P_Thrust(tmo,tangle,max(smo.scale,tthrust))
		if not tpain then P_Thrust(tmo,sangle+ANGLE_180,max(smo.scale,tthrust2)) end
	end
	
	//Get and apply zthrust
	if not(spain) then smo.momz = FixedMul(tmomz-smomz,tmo.scale*2-smo.scale) end
	if not(tpain) then tmo.momz = FixedMul(smomz-tmomz,smo.scale*2-tmo.scale) end
	
	//Knuckles interaction
	if smo.health and not(spain)
		and (splr.pflags&PF_GLIDING or splr.climbing 
			or (splr.charability == CA_GLIDEANDCLIMB) and collisiontype > 1 and P_IsObjectOnGround(smo) and not(splr.pflags&PF_SPINNING)
		) then
		splr.pflags = $&~(PF_GLIDING|PF_JUMPED)
		if not(P_IsObjectOnGround(smo)) or splr.climbing then
			splr.panim = PA_FALL
			splr.mo.state = S_PLAY_FALL
			splr.climbing = 0
			splr.lastsidehit = -1
			splr.lastlinehit = -1
		elseif (t_atk)
			splr.mo.state = S_PLAY_GLIDE_LANDING
			smo.momz = 0
		end
	end
	if tmo.health and not(tpain)
		and (tplr.pflags&PF_GLIDING or tplr.climbing 
			or (tplr.charability == CA_GLIDEANDCLIMB) and collisiontype > 1 and P_IsObjectOnGround(tmo) and not(tplr.pflags&PF_SPINNING)
		) then
		tplr.pflags = $&~(PF_GLIDING|PF_JUMPED)
		if not(P_IsObjectOnGround(tmo)) or tplr.climbing then
			tplr.panim = PA_FALL
			tplr.mo.state = S_PLAY_FALL
			tplr.climbing = 0
			tplr.lastsidehit = -1
			tplr.lastlinehit = -1
		elseif (s_atk)
			tplr.mo.state = S_PLAY_GLIDE_LANDING
		end
	end
	
	//Do recoil state
	if collisiontype > 1 then
		if not(spain) and (t_atk >= s_def)
			then
			//Uncurl
			splr.panim = PA_FALL
			smo.state = S_PLAY_FALL
			splr.pflags = $&~(PF_GLIDING|PF_JUMPED|PF_BOUNCING|PF_THOKKED|PF_SPINNING)
-- 			if (bias == -1 or FixedHypot(sspd,szspd) < FixedHypot(tspd,tzspd))
				//Apply recoil
				splr.powers[pw_nocontrol] = sthrust/smo.scale*2
				smo.recoilangle = sangle
				smo.recoilthrust = sthrust
-- 			end
		end
		if not(tpain) and (s_atk >= t_def)
			then
			tplr.panim = PA_FALL
			tmo.state = S_PLAY_FALL
			tplr.pflags = $&~(PF_GLIDING|PF_JUMPED|PF_BOUNCING|PF_THOKKED|PF_SPINNING)
-- 			if (bias == 1 or FixedHypot(sspd,szspd) > FixedHypot(tspd,tzspd))
				tplr.powers[pw_nocontrol] = tthrust/tmo.scale*2
				tmo.recoilangle = tangle
				tmo.recoilthrust = tthrust
-- 			end
		end
	end
	//Do Sound
	local defend = (s_def|t_def > 2)
	local attack = (s_atk|t_atk > 2)
	local damage = (tpain or spain)
	if not(defend or attack) then
		S_StartSound(smo,sfx_s3k5d) //Weak
	elseif not attack then
		S_StartSound(smo,sfx_s3k49) //Sturdy
	else
		S_StartSound(smo,sfx_s3k5f) //Powerful
	end

	//Player interactions completed
	return true
end

B.PlayerTouch = function(smo,tmo)
	if not(smo and smo.valid and tmo and tmo.valid and smo.player and tmo.player)
		or not(smo.health and tmo.health)
		or (smo.player.exiting or tmo.player.exiting)
		then return true end
	
	smo.player.homing = 0
	tmo.player.homing = 0
	
	if not(B.GetInteractionType(smo,tmo)) then
		if not(tmo.player.powers[pw_carry] or smo.player.powers[pw_carry] or P_PlayerInPain(tmo.player) or P_PlayerInPain(smo.player)) then		
			local sf = P_MobjFlip(smo)
			local tf = P_MobjFlip(tmo)
			if tmo.player.pflags&PF_CANCARRY/* and tmo.momz*tf > 0*/ and tmo.z*tf > (smo.z+smo.height/8*7)*sf then
				tmo.flags = $&~MF_SPECIAL			
				smo.flags = $&~MF_SPECIAL
			elseif smo.player.pflags&PF_CANCARRY /*and smo.momz*sf > 0*/ and smo.z*sf > (tmo.z+tmo.height/8*7)*tf then
				tmo.flags = $&~MF_SPECIAL			
				smo.flags = $&~MF_SPECIAL
			end
		end
	return true end //Don't bother trying to collide if nothing's going to come of it
	
	smo.pushed = tmo //'mo.pushed' will be referenced in ThinkFrame
	tmo.pushed = smo
	smo.flags = $&~MF_SPECIAL
	tmo.flags = $&~MF_SPECIAL
	tmo.player.pflags = $&~PF_CANCARRY
	smo.player.pflags = $&~PF_CANCARRY
	return true
end

B.UpdateCollisionHistory = function(pmo)
	if pmo.pushtics == nil then pmo.pushtics = 0 end //!! This shouldn't be necessary. Investigate later.
	if pmo.pushed then
		if pmo.pushtics == 0 then
			pmo.pushtics = 1
		else
			pmo.pushtics = $+1
			if pmo.pushtics > 15 then
				pmo.pushed_last = nil
				pmo.pushed = nil
				pmo.pushtics = 0
			end
		end
	else
		pmo.pushtics = 0
	end
	if pmo.pushed and pmo.pushed.valid and pmo.pushed.pushed == pmo and not(pmo.pushtics) then //Update other player object's pushed stats
		pmo.pushed.flags = $|MF_SPECIAL
		pmo.pushed.pushed_last = pmo
		pmo.pushed.pushed = nil
	end
	pmo.flags = $|MF_SPECIAL
	pmo.pushed_last = pmo.pushed
	pmo.pushed = nil
	local player = pmo.player
	if player.pushed_creditplr != nil
		if player.pushed_creditplr.valid and player.pushed_credittime then
			player.pushed_credittime = $-1
		else
			player.pushed_creditplr = nil
		end
	end
end