local B = CBW_Battle

B.TagGametype = function()
	if gametype == GT_EGGROBOTAG or G_TagGametype() then return true
	else return false end
end

B.BattleGametype = function()
	if gametype and B.Gametypes.Battle[gametype] then return true
	else return false end
end

B.CPGametype = function()
	if gametype and B.Gametypes.CP[gametype] then return true
	else return false end
end

B.TeamBattleScore = function()
	if not(G_GametypeHasTeams() and gametype and B.Gametypes.TeamScoreType[gametype] == 1) then return end
	for player in players.iterate()
		if player.scorelast == nil then
			player.scorelast = player.score
		end
		local score = max(0,player.score-player.scorelast)
		if player.ctfteam == 1 then
			redscore = $+score
		elseif player.ctfteam == 2 then
			bluescore = $+score
		end
		player.scorelast = player.score
	end
end

B.UpdatePlayerTrueMobjs = function()
	//Get counts
	local needtrue = 0
	for player in players.iterate() 
		if not(player.truemo) or not(player.truemo.valid) then
			player.truemo = nil
			needtrue = $+1
		end
	end
	//Iterate
	for mobj in mobjs.iterate()
		if mobj.player and not(mobj.player.truemo) then
			mobj.player.truemo = mobj
			needtrue = $-1
			if not(needtrue) then return true end
		end
	end
	return false
end

B.InitPlayer = function(player)
	//Action variables
	player.actiontype = 0
	player.actionrings = 0
	player.action2rings = 0
	player.actiontime = 0
	player.actionstate = 0
	player.actioncooldown = 2
	player.actionallowed = true
	player.actiontext = nil
	player.actiontextflags = 0
	player.action2text = nil
	player.action2textflags = 0
	player.charmed = false
	player.charmedtime = 0
	player.backdraft = 0
	player.spendrings = 0
	player.disableringslinger = false
	player.iseggrobo = false
-- 	player.willbeeggrobo = false
	player.capturing = false
	player.captureamount = 0
	player.gotflagdebuff = false
	player.pushed_creditplr = nil
	player.pushed_credittime = 0
	if player.mo then
		player.charflags = skins[player.mo.skin].flags
	end
	if player.autoaim == nil then
		player.autoaim = 1
	end
end

B.WaterFactor = function(mo)
	if mo.eflags&MFE_UNDERWATER then return 2 end
	return 1
end

B.UnderwaterMissile = function(mo)
	//Set underwater physics
	if mo.eflags&MFE_UNDERWATER and not(mo.waterslow) then
		mo.waterslow = true
		P_InstaThrust(mo,R_PointToAngle2(0,0,mo.momx,mo.momy),FixedHypot(mo.momx,mo.momy)/2)
		mo.momz = $/2
		if not(mo.staticfuse) then
			mo.fuse = $*2
		end
	end
	//Cancel underwater slowdown
	if not(mo.eflags&MFE_UNDERWATER) and(mo.waterslow) then
		mo.waterslow = false
		P_InstaThrust(mo,R_PointToAngle2(0,0,mo.momx,mo.momy),FixedHypot(mo.momx,mo.momy)*2)
		mo.momz = $*2
		if not(mo.staticfuse) then
			mo.fuse = $/2
		end
	end
end

B.AttackText = function(player)
	local pflags = player.pflags
	local ability = player.charability
	local ability2 = player.charability2
	local actiontype = player.actiontype
	local actionstate = player.actionstate
	local sflags = player.charflags
	//Priority
	if player.powers[pw_invulnerability] or player.powers[pw_super] then return end
	if (player.powers[pw_shield] == SH_ARMAGEDDON) and pflags&PF_SHIELDABILITY then return end
	//Ability specific
	if player.powers[pw_shield] == SH_ATTRACT and pflags&PF_SHIELDABILITY then return "homing attack" end
	if actiontype == CA_THOK and actionstate == 1 then return "ground pound" end
	if actiontype == CA_THOK and actionstate == 2 then return "super jump" end
	if ability == CA_THOK and pflags&PF_THOKKED and not(pflags&PF_SHIELDABILITY) then return "speed thok" end
	if ability == CA_FLY and pflags&PF_THOKKED and not(pflags&PF_SHIELDABILITY) then return "propeller attack" end
	if actiontype == CA_GLIDEANDCLIMB and actionstate == 2 then return "drill dive" end
	if ability == CA_GLIDEANDCLIMB and pflags&PF_GLIDING then return "gliding fists" end
	if ability == CA_TWINSPIN and player.panim == PA_ABILITY then return "twin spin attack" end
	if ability2 == CA2_MELEE and player.panim == PA_ABILITY2 then return "hammer thrust" end
	if ability == CA_BOUNCE and PF_BOUNCING then return "tail bounce" end
	if player.dashmode > TICRATE*3 then return "dash attack" end
	//General
	if pflags&PF_JUMPED and not(pflags&PF_NOJUMPDAMAGE) then return "jumping spin attack" end
	if pflags&PF_SPINNING then return "spin attack" end
	if pflags&PF_JUMPED and sflags&SF_STOMPDAMAGE then return "jump attack" end
	//Default
	return "attack"
end

B.CustomHurtMessage = function(player,source,bullettext,hurttext,killtext)
	//Defining playertext
	local playertext = player.name
	if G_GametypeHasTeams() then
		if player.ctfteam == 1 then playertext = "\x85"..$.."\x80" end
		if player.ctfteam == 2 then playertext = "\x84"..$.."\x80" end
	end
	//Defining sourcetext
	local sourcetext
	if source and source.player then 
		sourcetext = source.player.name
		if G_GametypeHasTeams() then
			if source.player.ctfteam == 1 then sourcetext = "\x85"..$.."\x80" end
			if source.player.ctfteam == 2 then sourcetext = "\x84"..$.."\x80" end
		end
	end
	//Defining damagetext
	if hurttext == nil then hurttext = "hit" end
	if killtext == nil then killtext = "killed" end
	//Which do we choose?
	local choosehurttext
	if (player.mo.health) then
		choosehurttext = hurttext
	else
		choosehurttext = killtext
	end
	local text = ""
	//If no sourceplayer
	if source == nil then
		text = playertext.." was "..choosehurttext
		if (bullettext) then
			text = $.." by "..bullettext.."."
		end
	//If attacking player
	else
		text = sourcetext
		if (bullettext) then
			text = $.."'s "..bullettext
		end	
		text = $.." "..choosehurttext.." "..playertext.."."
	end
	return text
end

B.PlayerButtonPressed = function(player,button,held,stasis)
		if not(player.cmd.buttons&button) then return false end
		if held == true and not(player.buttonhistory&button) then return false end
		if held == false and player.buttonhistory&button then return false end
		if(stasis) then return false end
		return true
end

B.PlayerCanBeDamaged = function(player)
	if not(player.playerstate == PST_LIVE) then return false end
	if player.powers[pw_invulnerability] then return false end
	if player.powers[pw_super] then return false end
	if player.powers[pw_flashing] then return false end
	return true
end

B.MyTeam = function(player,myplayer)
	//Check yourself before you wreck yourself
	if myplayer == player then return true end
	//FriendlyFire
	if CV_FindVar("friendlyfire").value then
		return false
	end
	//CTF checks
	if G_GametypeHasTeams() then
		if player.ctfteam == myplayer.ctfteam then return true
		else return false
		end
	end
	//Tag checks
	if B.TagGametype() then
		if player.pflags&PF_TAGIT == myplayer.pflags&PF_TAGIT then return true
		else return false
		end
	end
	//Battle check
	if (B.BattleGametype() or G_RingSlingerGametype()) and not(G_GametypeHasTeams()) then
		return false
	end
	//default
	return true
end

B.GetNearestPlayer = function(mo,fov,teamcheck,aim,ignorepain)
	if not(mo.player) and not (teamcheck == nil or teamcheck == 0) then return nil end //Teamchecks aren't allowed on nonplayer objects
	local nearest
	local neardist
	//Cycle through players
	for player in players.iterate
		//Failchecks
		if not (player.mo and player.playerstate == PST_LIVE) then continue end //Invalid player object
		if player.mo == mo then continue end // Don't count ourselves!
		if (teamcheck == 1 and not(B.MyTeam(player,mo.player))) //Need teammate
			or (teamcheck == -1 and B.MyTeam(player,mo.player)) //Need opponent
			or not(P_CheckSight(player.mo,mo)) //Not in view
			or player.mo.flags&MF_NOCLIPTHING //Can't target!
			then continue
		end
		if ignorepain == true and not(B.PlayerCanBeDamaged(player)) then continue end //Ignore invulnerable players
		if fov != nil then
			if aim == nil then aim = mo.angle end
			//Get angle
			local ang = R_PointToAngle2(mo.x,mo.y,player.mo.x,player.mo.y)-aim
			if ang > 180*ANG1 then ang = $-360*ANG1
			elseif ang < -180*ANG1 then ang = $+360*ANG1
			end
			//Check field of view
			if (ang < -fov/2*ANG1 or ang > fov/2*ANG1) then continue end
		end
		//Get distance
		local dist = R_PointToDist2(player.mo.x,player.mo.y,mo.x,mo.y)
		if nearest == nil or neardist > dist then
			nearest = player
			neardist = dist
		end
	end
	return nearest
end

B.InstaThrustZAim = function(mo,xyangle,zangle,speed)
	local xythrust = P_ReturnThrustX(nil,zangle,speed)
	local zthrust = P_ReturnThrustY(nil,zangle,speed)
	P_InstaThrust(mo,xyangle,xythrust)		
	mo.momz = $+zthrust	
end

B.SearchObject = function(mo,aim,fov,range,searchflags,refmo,teamcheck)
		if refmo == nil then refmo = mo end
		local team
		if not(teamcheck == nil or teamcheck == 0) then
			if (mo.player) then
				team = mo.player.ctfteam
			else
				team = mo.ctfteam
			end
			if team == nil or team == 0 then
				team = 0
				teamcheck = 0
			end
		end
		local m = nil
		searchBlockmap('objects',function(mo,found)
			local dist = R_PointToAngle2(mo.x,mo.y,found.x,found.y)
			local angle = abs(dist-aim)
			if found != refmo and found.flags&searchflags and P_CheckSight(mo,found) and found.health and abs(angle) < fov/2*ANG1
				and (
					(teamcheck == nil or teamcheck == 0) or (teamcheck == 1 and team == found.ctfteam) or (teamcheck == -1 and team != found.ctfteam)
				)
				and
					not(refmo and refmo.type == found.type and refmo.target == found.target)
				and (
					not(m and m.valid) 
					or (R_PointToDist2(mo.x,mo.y,m.x,m.y)+abs(mo.z-m.z) > R_PointToDist(mo.x,mo.y,found.x,found.y)+abs(mo.z-found.z))
				) then
				m = found
			end
		end,mo,mo.x-range,mo.x+range,mo.y-range,mo.y+range)
		return m
end

B.AutoAim = function(mo,aim,fov,teamcheck,ignorepain,projectile,speed,searchflags)
	local p = B.GetNearestPlayer(mo,fov,teamcheck,aim,ignorepain)
	if p then 
		local dist = FixedHypot(FixedHypot(p.mo.momx,p.mo.momy),p.mo.momz)
		local x1,y1,z1 = projectile.x,projectile.y,projectile.z
		local x2,y2,z2 = p.mo.x+p.mo.momx,p.mo.y+p.mo.momy,p.mo.z+p.mo.height/2+p.mo.momz
		local zang = B.GetZAngle(x1,y1,z1,x2,y2,z2)
		local xyang = R_PointToAngle2(x1,y1,x2,y2)
		B.InstaThrustZAim(projectile,xyang,zang,speed)
		return p.mo
	else
		local p = B.SearchObject(mo,aim,fov,mo.scale*640,searchflags,nil,teamcheck)
		if p then
			local x1,y1,z1 = projectile.x,projectile.y,projectile.z
			local x2,y2,z2 = p.x+p.momx,p.y+p.momy,p.z+p.momz
			local zang = max(-ANG60,min(ANG60,B.GetZAngle(projectile.x,projectile.y,projectile.z,p.x,p.y,p.z+projectile.height/2)))
			local xyang = R_PointToAngle2(x1,y1,x2,y2)
			B.InstaThrustZAim(projectile,xyang,zang,speed)
			return p
		end
	end
	return false
end

B.InstaThrustSpread = function(mo,aim,spread,angle,speed)
	local xyangle = (angle-90)*ANG1*2
	local zangle = angle*ANG1*2
	local xymul = FRACUNIT/180*spread
	local zmul = FRACUNIT/180*spread
	if sin(angle*ANG1) < 0 then zmul = $*-1 end
	if cos(angle*2*ANG1) < 0 then xymul = -1*zmul end
	xyangle = FixedMul($,xymul)
	zangle = FixedMul(abs($),zmul)
	B.InstaThrustZAim(mo,xyangle+aim,zangle,speed)
end

B.ResetPlayerProperties = function(player,jumped,thokked)
	local mo = player.mo
	if not(mo) then return end
	//Reset pflags
	local pflags = player.pflags&~(PF_GLIDING|PF_BOUNCING)
	if jumped == true and not(mo.state == S_PLAY_SPRING) then
		pflags = $|PF_JUMPED&~(PF_STARTJUMP|PF_NOJUMPDAMAGE)
		mo.state = S_PLAY_ROLL
	elseif jumped == false then
		pflags = $&~(PF_JUMPED|PF_SPINNING)
		if not(P_IsObjectOnGround(mo)) and not(P_PlayerInPain(player)) then
			mo.state = S_PLAY_FALL
		end
	end
	if thokked == true then pflags = $|PF_THOKKED
	elseif thokked == false then pflags = $&~PF_THOKKED
	end
	player.pflags = pflags
	//Reset other variables
	local skin = skins[mo.skin]
	mo.flags = $&~(MF_NOCLIPTHING)
	mo.flags2 = $&~(MF2_DONTDRAW)
	player.charability = skin.ability
	player.charability2 = skin.ability2
	player.normalspeed = skin.normalspeed
	player.thrustfactor = skin.thrustfactor
	player.climbing = 0
	player.secondjump = 0
	player.actionstate = 0
	player.actiontime = 0
	if mo.rollangle != 0 then 
		mo.rollangle = 0
		mo.player.drawangle = mo.angle
	end
end

B.OverlayHide = function(mo,owner)
	if not(owner and owner.valid) then return nil end
	if owner.flags&MF_NOCLIPTHING or owner.flags2&MF2_DONTDRAW then mo.flags2 = $|MF2_DONTDRAW return true
	else mo.flags2 = $&~MF2_DONTDRAW return false
	end
end

B.BattleProjectileDamage = function(target,inflictor,source,something,damage)
	if not(B.BattleGametype()) then return nil end
	if not(inflictor and inflictor.valid) then return nil end
	if not(inflictor.flags&(MF_MISSILE|MF_ENEMY)) then return nil end
	if not(source and source.player and target.player) then return nil end
	if (target.player.pflags&PF_TAGIT) then return false end

	if not(B.PlayerCanBeDamaged(target.player) or inflictor.flags2&MF2_SUPERFIRE) then return nil end
	local element = damage&~(DMG_CANHURTSELF)
	if (element == DMG_WATER and target.player.powers[pw_shield]&SH_PROTECTWATER)
		or (element == DMG_FIRE and target.player.powers[pw_shield]&SH_PROTECTFIRE)
		or (element == DMG_ELECTRIC and target.player.powers[pw_shield]&SH_PROTECTELECTRIC)
		or (element == DMG_SPIKE and target.player.powers[pw_shield]&SH_PROTECTSPIKE)
		then
		return nil
	end
	if (gametype == GT_EGGROBOTAG and source.player.pflags&PF_TAGIT) and not(target.player.pflags&PF_TAGIT)
		then
		print(target.player.name.." has been found!")
		target.player.pflags = $|PF_TAGIT
		return true
	end
	if not(B.MyTeam(target.player,source.player))
		return true
	else
		return nil
	end
end