local B = CBW_Battle
local PFunc = B.PriorityFunction

B.InitPriority = function(player)
	player.battle_func = nil //Index for a function that handles core priority behavior, like spindash/spinjump priority
	player.battle_extfunc = nil //Priority type script index for certain character-specific abilities
	B.SetPlayerPriorityFunc(player) //Overrides the above functions depending on the player skin
	player.battle_atk = 0 //Current attack priority
	player.battle_def = 0 //Current defense priority. If higher than enemy attack priority, player will not take damage. If equal, player will be "uncurled".
	player.battle_sfunc = nil //Index for a function returning whether a character has hit a "sweetspot" or "sourspot". Function allows arg1(player.mo) and arg2(otherplayer.mo). If true, battle_satk and battle_sdef override battle_atk and battle_def.
	player.battle_satk = 0 //Sweetspot/sourspot attack priority
	player.battle_sdef = 0 //Sweetspot/sourspot defense priority
end

B.SetPriority = function(player,atk,def,sfunc,satk,sdef)
	if atk != nil then player.battle_atk = atk end
	if def != nil then player.battle_def = def end
	if sfunc != nil then player.battle_sfunc = sfunc else player.battle_sfunc = nil end
	if satk != nil then player.battle_satk = satk end
	if sdef != nil then player.battle_sdef = sdef end
end

local xyangle = B.GetCollideRelativeAngle
local zangle = B.GetZCollideAngle

B.SPriority_Fly = function(mo,other)
	if zangle(mo,other) > -ANGLE_45 then return false
	else return true end
end

B.SPriority_Glide = function(mo,other)
	if abs(zangle(mo,other)) > ANGLE_45 or abs(xyangle(mo,other,true)) > ANGLE_45 then return false
	else return true end
end

B.SPriority_TwinSpin = function(mo,other)
	if mo.player then mo.player.pflags = $&~PF_THOKKED end
	if xyangle(mo,other,true) > ANGLE_45 then return false
	else return true end
end

B.SPriority_Melee = function(mo,other)
	if abs(xyangle(mo,other,true)) > ANGLE_45 then return false
	else return true end
end

B.SPriority_TailBounce = function(mo,other)
	if (B.GetZCollideAngle(mo,other) < ANGLE_45) then return false
	else return true end
end

B.SPriority_StompDamage = function(mo,other)
	local bottomheight = mo.z
	local topheight = mo.z + mo.height

	if (mo.eflags & MFE_VERTICALFLIP)
		local swap = bottomheight
		bottomheight = topheight
		topheight = swap
	end

	if not ((P_MobjFlip(mo)*(bottomheight - (other.z + other.height/2)) > 0)
	and (P_MobjFlip(mo)*(mo.momz - other.momz) < 0)) then return false
	else return true end
end

B.Priority_Core = function(player)
	local pflags = player.pflags
-- 	local shield = pflags&PF_SHIELDABILITY
	local spinjump = (pflags&PF_JUMPED and not(pflags&PF_NOJUMPDAMAGE))
	local spinning = pflags&PF_SPINNING
	local stomping = (player.mo and player.mo.valid
	and (skins[player.mo.skin].flags&SF_STOMPDAMAGE) and not P_PlayerInPain(player))
	
	//Default
	player.battle_def = 0
	player.battle_atk = 0
	player.battle_sfunc = nil
	player.battle_satk = 0
	player.battle_sdef = 0
		
	//Spin attack and Stomp Damage
	if spinjump or spinning then
		B.SetPriority(player,1,1,nil,1,1)
	elseif stomping
		B.SetPriority(player,0,0,"stomp_damage",1,1)
	end
end

B.Priority_Ability = function(player)
	local grounded = P_IsObjectOnGround(player.mo)
	local pflags = player.pflags
	local abil1 = player.charability
	local abil2 = player.charability2
	local anim1 = (player.panim == PA_ABILITY)
	local anim2 = (player.panim == PA_ABILITY2)
	local thokked = pflags&PF_THOKKED
	local shield = pflags&PF_SHIELDABILITY
	local spinjump = (pflags&PF_JUMPED and not(pflags&PF_NOJUMPDAMAGE))
	local spinning = pflags&PF_SPINNING
	local knuckles = (abil1 == CA_GLIDEANDCLIMB)
	local flying = (abil1 ==CA_FLY and thokked&~shield)
	local gliding = pflags&PF_GLIDING
	local twinspin = (abil1 ==CA_TWINSPIN and anim1)
	local melee = (abil2 ==CA2_MELEE and anim2)
	local tailbounce = pflags&PF_BOUNCING
	local dashing = player.dashmode > 3*TICRATE and not(player.pflags&PF_STARTDASH)
	local prepdash = player.dashmode > 3*TICRATE and player.pflags&PF_STARTDASH
	local actiontype = B.ActionChar[player.actiontype]
	
	//Knuckles Guard
	if knuckles and not(spinning or spinjump) then
		B.SetPriority(player,0,0,"knuckles_glide",0,1)
	end
	
	//Sonic
	if actiontype == B.Action.SuperSpinJump and player.actionstate and spinjump then
		B.SetPriority(player,1,2,nil,1,2)
	end
	
	//Tails
	if flying then
		B.SetPriority(player,0,0,"tails_fly",2,2)
	end
	//Knuckles
	if gliding then
		B.SetPriority(player,1,0,"knuckles_glide",2,1)
	end
	//Amy
	if twinspin then 
		B.SetPriority(player,1,2,"amy_twinspin",2,3)
	end
	if melee then 
		B.SetPriority(player,1,0,"amy_melee",1,3)
	end
	//Fang
	if tailbounce then
		B.SetPriority(player,0,0,"fang_tailbounce",2,3)
	end
	//Metal
	if dashing then
		B.SetPriority(player,3,1,nil,3,1)
	elseif prepdash then
		B.SetPriority(player,1,3,nil,1,3)
	end
end

//Do frame priority
B.DoPriority = function(player)
	local func = player.battle_func
	local func2 = player.battle_extfunc
	if func and PFunc[func] then
		PFunc[func](player)
	end
	if func2 and PFunc[func2] then
		PFunc[func2](player)
	end
end

//Add sweetspot/sourspot priority
B.DoSPriority = function(player,othermo)
	local func = player.battle_sfunc
	if not(func and PFunc[func]) then return end
	
	if PFunc[func](player.mo,othermo) then
		player.battle_atk = player.battle_satk
		player.battle_def = player.battle_sdef
	end
end